@@ -2,13 +2,21 @@
 #this repository contains the full copyright notices and license terms.
 import datetime
 from decimal import Decimal
+from operator import itemgetter
+from sql import Literal, Union, Column
+from sql.aggregate import Max, Sum
+from sql.functions import Now
+from sql.conditionals import Coalesce
+
 from trytond.model import ModelSQL, ModelView, fields
 from trytond.wizard import Wizard, StateView, StateAction, Button
 from trytond.pyson import PYSONEncoder, Eval, Or
 from trytond.transaction import Transaction
-from trytond.tools import safe_eval, reduce_ids
+from trytond.tools import reduce_ids
 from trytond.pool import Pool, PoolMeta
 
+from .move import StockMixin
+
 __all__ = ['Template', 'Product',
     'ProductByLocationStart', 'ProductByLocation',
     'ProductQuantitiesByWarehouse', 'ProductQuantitiesByWarehouseStart',
@@ -69,7 +77,8 @@ class Template:
         super(Template, cls).write(templates, vals)
 
 
-class Product:
+class Product(object, StockMixin):
+    __metaclass__ = PoolMeta
     __name__ = "product.product"
     quantity = fields.Function(fields.Float('Quantity'), 'get_quantity',
             searcher='search_quantity')
@@ -80,87 +89,20 @@ class Product:
 
     @classmethod
     def get_quantity(cls, products, name):
-        Date = Pool().get('ir.date')
-
-        quantities = dict((p.id, 0.0) for p in products)
-        if not Transaction().context.get('locations'):
-            return quantities
-
-        context = {}
-        if (name == 'quantity'
-                and Transaction().context.get('stock_date_end')
-                and Transaction().context.get('stock_date_end') >
-                Date.today()):
-            context['stock_date_end'] = Date.today()
-
-        if name == 'forecast_quantity':
-            context['forecast'] = True
-            if not Transaction().context.get('stock_date_end'):
-                context['stock_date_end'] = datetime.date.max
-        with Transaction().set_context(context):
-            pbl = cls.products_by_location(
-                location_ids=Transaction().context['locations'],
-                product_ids=quantities.keys(), with_childs=True)
-
-        for location in Transaction().context['locations']:
-            for product in products:
-                quantities[product.id] += pbl.get((location, product.id), 0.0)
-        return quantities
-
-    @staticmethod
-    def _search_quantity_eval_domain(line, domain):
-        field, operator, operand = domain
-        value = line.get(field)
-        if value is None:
-            return False
-        if operator not in ("=", ">=", "<=", ">", "<", "!="):
-            return False
-        if operator == "=":
-            operator = "=="
-        return (safe_eval(str(value) + operator + str(operand)))
+        location_ids = Transaction().context.get('locations')
+        return cls._get_quantity(products, name, location_ids, products)
 
     @classmethod
     def search_quantity(cls, name, domain=None):
-        Date = Pool().get('ir.date')
-
-        if not (Transaction().context.get('locations') and domain):
-            return []
-
-        context = {}
-        if (name == 'quantity'
-                and Transaction().context.get('stock_date_end')
-                and Transaction().context.get('stock_date_end') >
-                Date.today()):
-            context['stock_date_end'] = Date.today()
-
-        if name == 'forecast_quantity':
-            context['forecast'] = True
-            if not Transaction().context.get('stock_date_end'):
-                context['stock_date_end'] = datetime.date.max
-
-        with Transaction().set_context(context):
-            pbl = cls.products_by_location(
-                    location_ids=Transaction().context['locations'],
-                    with_childs=True, skip_zero=False).iteritems()
-
-        processed_lines = []
-        for (location, product), quantity in pbl:
-            processed_lines.append({
-                    'location': location,  # XXX useful ?
-                    'product': product,
-                    name: quantity,
-                    })
-
-        res = [line['product'] for line in processed_lines
-            if cls._search_quantity_eval_domain(line, domain)]
-        return [('id', 'in', res)]
+        location_ids = Transaction().context.get('locations')
+        return cls._search_quantity(name, location_ids, domain)
 
     @classmethod
     def get_cost_value(cls, products, name):
         cost_values = {}
         context = {}
         trans_context = Transaction().context
-        if trans_context.get('stock_date_end'):
+        if 'stock_date_end' in context:
             context['_datetime'] = trans_context['stock_date_end']
         with Transaction().set_context(context):
             for product in products:
@@ -174,7 +116,7 @@ class Product:
 
     @classmethod
     def products_by_location(cls, location_ids, product_ids=None,
-            with_childs=False, skip_zero=True):
+            with_childs=False, grouping=('product',)):
         """
         Compute for each location and product the stock quantity in the default
         uom of the product.
@@ -193,9 +135,9 @@ class Product:
                     the storage zone.
         If product_ids is None all products are used.
         If with_childs, it computes also for child locations.
-        If skip_zero it lists also items with zero quantity
+        grouping defines how stock moves are grouped.
 
-        Return a dictionary with (location id, product id) as key
+        Return a dictionary with location id and grouping as key
                 and quantity as value.
         """
         pool = Pool()
@@ -204,6 +146,13 @@ class Product:
         Location = pool.get('stock.location')
         Date = pool.get('ir.date')
         Period = pool.get('stock.period')
+        Move = pool.get('stock.move')
+        Product = pool.get('product.product')
+        Template = pool.get('product.template')
+
+        move = Move.__table__()
+        product = Product.__table__()
+        template = Template.__table__()
 
         today = Date.today()
 
@@ -211,6 +160,12 @@ class Product:
             return {}
         cursor = Transaction().cursor
         context = Transaction().context.copy()
+
+        for field in grouping:
+            if field not in Move._fields:
+                raise ValueError('"%s" has no field "%s"' % (Move, field))
+        assert 'product' in grouping
+
         # Skip warehouse location in favor of their storage location
         # to compute quantities. Keep track of which ids to remove
         # and to add after the query.
@@ -227,9 +182,14 @@ class Product:
                 wh_to_add[location.id] = location.storage_location.id
         location_ids = list(location_ids)
 
-        move_rule_query, move_rule_val = Rule.domain_get('stock.move')
+        move_rule_query = Rule.domain_get('stock.move')
+        if move_rule_query is None:
+            move_rule_query = Literal(True)
 
-        period_clause, period_vals = 'period = %s', [0]
+        PeriodCache = Period.get_cache(grouping)
+        period = None
+        if PeriodCache:
+            period_cache = PeriodCache.__table__()
 
         if not context.get('stock_date_end'):
             context['stock_date_end'] = datetime.date.max
@@ -239,152 +199,105 @@ class Product:
                 or (context['stock_date_end'] == today
                     and not context.get('forecast'))):
             state_date_clause = (
-                '('
-                    '(state in (%s, %s)) '
-                'AND '
-                    '('
-                        '('
-                            '(effective_date IS NULL) '
-                        'AND '
-                            '(planned_date <= %s) '
-                        ') '
-                    'OR '
-                        '(effective_date <= %s)'
-                    ')'
-                ')')
-            state_date_vals = ["done",
-                    context.get('stock_assign') and 'assigned' or 'done',
-                    context['stock_date_end'],
-                    context['stock_date_end'],
-                    ]
+                move.state.in_(['done',
+                        context.get('stock_assign') and 'assigned' or 'done'])
+                & (
+                    (
+                        (move.effective_date == None)
+                        & (move.planned_date <= context['stock_date_end'])
+                        )
+                    | (move.effective_date <= context['stock_date_end'])
+                    )
+                )
         # future date end: filter move on state done and date
         # before today, or on all state and date between today and
         # date_end.
         else:
             state_date_clause = (
-                '('
-                    '('
-                        '(state in (%s, %s)) '
-                    'AND '
-                        '('
-                            '('
-                                '(effective_date IS NULL) '
-                            'AND '
-                                '(planned_date <= %s) '
-                            ') '
-                        'OR '
-                            '(effective_date <= %s)'
-                        ')'
-                    ')'
-                'OR '
-                    '('
-                        '(state in (%s, %s, %s)) '
-                    'AND '
-                        '('
-                            '('
-                                '(effective_date IS NULL) '
-                            'AND '
-                                '(COALESCE(planned_date, %s) <= %s) '
-                            'AND '
-                                '(COALESCE(planned_date, %s) >= %s)'
-                            ')'
-                        'OR '
-                            '('
-                                '(effective_date <= %s) '
-                            'AND '
-                                '(effective_date >= %s)'
-                            ')'
-                        ')'
-                    ')'
-                ')')
-
-            state_date_vals = [
-                'done', context.get('stock_assign') and 'assigned' or 'done',
-                today, today,
-                'done', 'assigned', 'draft',
-                datetime.date.max, context['stock_date_end'],
-                datetime.date.max, today,
-                context['stock_date_end'], today,
-                ]
+                (move.state.in_(['done',
+                            context.get('stock_assign') and 'assigned'
+                            or 'done'])
+                    & (
+                        (
+                            (move.effective_date == None)
+                            & (move.planned_date <= today)
+                            )
+                        | (move.effective_date <= today)
+                        )
+                    )
+                | (move.state.in_(['done', 'assigned', 'draft'])
+                    & (
+                        (
+                            (move.effective_date == None)
+                            & (Coalesce(move.planned_date, datetime.date.max)
+                                <= context['stock_date_end'])
+                            & (Coalesce(move.planned_date, datetime.date.max)
+                                >= today)
+                            )
+                        | (
+                            (move.effective_date <= context['stock_date_end'])
+                            & (move.effective_date >= today)
+                            )
+                        )
+                    )
+                )
 
         if context.get('stock_date_start'):
             if context['stock_date_start'] > today:
-                state_date_clause += ('AND '
-                    '('
-                        '(state in (%s, %s, %s)) '
-                    'AND '
-                        '('
-                            '('
-                                '(effective_date IS NULL) '
-                            'AND '
-                                '('
-                                    '(planned_date >= %s) '
-                                'OR '
-                                    '(planned_date IS NULL)'
-                                ')'
-                            ') '
-                        'OR '
-                            '(effective_date >= %s)'
-                        ')'
-                    ')')
-                state_date_vals.extend(['done', 'assigned', 'draft',
-                     context['stock_date_start'], context['stock_date_start']])
+                state_date_clause &= (
+                    move.state.in_(['done', 'assigned', 'draft'])
+                    & (
+                        (
+                            (move.effective_date == None)
+                            & (
+                                (move.planned_date >=
+                                    context['stock_date_start'])
+                                | (move.planned_date == None)
+                                )
+                            )
+                        | (move.effective_date >= context['stock_date_start'])
+                        )
+                    )
             else:
-                state_date_clause += ('AND '
-                    '('
-                        '('
-                            '(state in (%s, %s, %s)) '
-                        'AND '
-                            '('
-                                '('
-                                    '(effective_date IS NULL) '
-                                'AND '
-                                    '('
-                                        '(planned_date >= %s) '
-                                    'OR '
-                                        '(planned_date IS NULL)'
-                                    ') '
-                                ')'
-                            'OR '
-                                '(effective_date >= %s)'
-                            ')'
-                        ') '
-                    'OR '
-                        '('
-                            '(state in (%s, %s)) '
-                        'AND '
-                            '('
-                                '('
-                                    '(effective_date IS NULL) '
-                                'AND '
-                                    '('
-                                        '('
-                                            '(planned_date >= %s) '
-                                        'AND '
-                                            '(planned_date < %s)'
-                                        ') '
-                                    'OR '
-                                        '(planned_date IS NULL)'
-                                    ')'
-                                ') '
-                            'OR '
-                                '('
-                                    '(effective_date >= %s) '
-                                'AND '
-                                    '(effective_date < %s)'
-                                ')'
-                            ')'
-                        ')'
-                    ')')
-
-                state_date_vals.extend(['done', 'assigned', 'draft',
-                    today, today,
-                    'done',
-                    context.get('stock_assign') and 'assigned' or 'done',
-                    context['stock_date_start'], today,
-                    context['stock_date_start'], today,
-                    ])
-        else:
+                state_date_clause &= (
+                    (
+                        move.state.in_(['done', 'assigned', 'draft'])
+                        & (
+                            (
+                                (move.effective_date == None)
+                                & (
+                                    (move.planned_date >= today)
+                                    | (move.planned_date == None)
+                                    )
+                                )
+                            | (move.effective_date >= today)
+                            )
+                        )
+                    | (
+                        move.state.in_(['done',
+                                context.get('stock_assign') and 'assigned'
+                                or 'done'])
+                        & (
+                            (
+                                (move.effective_date == None)
+                                & (
+                                    (
+                                        (move.planned_date >=
+                                            context['stock_date_start'])
+                                        & (move.planned_date < today)
+                                        )
+                                    | (move.planned_date == None)
+                                    )
+                                )
+                            | (
+                                (move.effective_date >=
+                                    context['stock_date_start'])
+                                & (move.effective_date < today)
+                                )
+                            )
+                        )
+                    )
+        elif PeriodCache:
             with Transaction().set_user(0, set_context=True):
                 periods = Period.search([
                         ('date', '<', context['stock_date_end']),
@@ -392,105 +305,104 @@ class Product:
                         ], order=[('date', 'DESC')], limit=1)
             if periods:
                 period, = periods
-                state_date_clause += (' AND '
-                    '(COALESCE(effective_date, planned_date, %s) > %s)')
-                state_date_vals.extend([datetime.date.max, period.date])
-                period_vals[0] = period.id
+                state_date_clause &= (
+                    Coalesce(move.effective_date, move.planned_date,
+                        datetime.date.max) > period.date)
 
         if with_childs:
-            query, args = Location.search([
+            location_query = Location.search([
                     ('parent', 'child_of', location_ids),
-                    ], query_string=True, order=[])
-            where_clause = " IN (" + query + ") "
-            where_vals = args
+                    ], query=True, order=[])
         else:
-            where_clause = " IN (" + \
-                ",".join(('%s',) * len(location_ids)) + ") "
-            where_vals = location_ids[:]
+            location_query = location_ids[:]
 
-        if move_rule_query:
-            move_rule_query = " AND " + move_rule_query + " "
-
-        product_template_join = ""
-        product_template_join_period = ""
+        from_ = move
+        if PeriodCache:
+            from_period = period_cache
         if product_ids:
-            red_clause, red_vals = reduce_ids('product', product_ids)
-            where_clause += "AND " + red_clause
-            where_vals += red_vals
+            where = reduce_ids(move.product, product_ids)
+            if PeriodCache:
+                where_period = reduce_ids(period_cache.product, product_ids)
         else:
-            where_clause += "AND product_template.active = %s"
-            where_vals.append(True)
-            product_template_join = (
-                "JOIN product_product "
-                    "ON (stock_move.product = product_product.id) "
-                "JOIN product_template "
-                    "ON (product_product.template = "
-                        "product_template.id) ")
-            product_template_join_period = (
-                "JOIN product_product "
-                    "ON (stock_period_cache.product = product_product.id) "
-                "JOIN product_template "
-                    "ON (product_product.template = product_template.id) ")
+            where = where_period = template.active == True
+            from_ = from_.join(product, condition=move.product == product.id)
+            from_ = from_.join(template,
+                condition=product.template == template.id)
+            if PeriodCache:
+                from_period = from_period.join(product,
+                    condition=period_cache.product == product.id)
+                from_period = from_period.join(template,
+                    condition=product.template == template.id)
 
         if context.get('stock_destinations'):
             destinations = context.get('stock_destinations')
-            dest_clause_from = " AND from_location in ("
-            dest_clause_from += ",".join(('%s',) * len(destinations))
-            dest_clause_from += ") "
-            dest_clause_to = " AND to_location in ("
-            dest_clause_to += ",".join(('%s',) * len(destinations))
-            dest_clause_to += ") "
-            dest_vals = destinations
+            dest_clause_from = move.from_location.in_(destinations)
+            dest_clause_to = move.to_location.in_(destinations)
 
-            dest_clause_period = (' AND location IN ('
-                + ','.join(('%s',) * len(destinations)) + ') ')
+            if PeriodCache:
+                dest_clause_period = period_cache.location.in_(destinations)
 
         else:
-            dest_clause_from = dest_clause_to = dest_clause_period = ""
-            dest_vals = []
+            dest_clause_from = dest_clause_to = dest_clause_period = \
+                Literal(True)
 
         # The main select clause is a union between three similar subqueries.
         # One that sums incoming moves towards locations, one that sums
         # outgoing moves and one for the period cache.  UNION ALL is used
         # because we already know that there will be no duplicates.
-        select_clause = (
-                "SELECT location, product, sum(quantity) AS quantity "
-                "FROM ( "
-                    "SELECT to_location AS location, product, "
-                        "SUM(internal_quantity) AS quantity "
-                    "FROM stock_move " + product_template_join + " "
-                    "WHERE (%s) "
-                        "AND to_location %s "
-                    "GROUP BY to_location, product "
-                    "UNION ALL "
-                    "SELECT from_location AS location, product, "
-                        "-SUM(internal_quantity) AS quantity "
-                    "FROM stock_move " + product_template_join + " "
-                    "WHERE (%s) "
-                        "AND from_location %s "
-                    "GROUP BY from_location, product, uom "
-                    "UNION ALL "
-                    "SELECT location, product, internal_quantity AS quantity "
-                    "FROM stock_period_cache "
-                        + product_template_join_period + " "
-                    "WHERE (%s) "
-                        "AND location %s "
-                ") AS T GROUP BY T.location, T.product")
-
-        cursor.execute(select_clause % (
-            state_date_clause,
-            where_clause + move_rule_query + dest_clause_from,
-            state_date_clause,
-            where_clause + move_rule_query + dest_clause_to,
-            period_clause, where_clause + dest_clause_period),
-            state_date_vals + where_vals + move_rule_val + dest_vals +
-            state_date_vals + where_vals + move_rule_val + dest_vals +
-            period_vals + where_vals + dest_vals)
+        move_keys = [Column(move, key).as_(key) for key in grouping]
+        query = from_.select(move.to_location.as_('location'),
+            Sum(move.internal_quantity).as_('quantity'),
+            *move_keys,
+            where=state_date_clause
+            & where
+            & move.to_location.in_(location_query)
+            & move.id.in_(move_rule_query)
+            & dest_clause_from,
+            group_by=[move.to_location] + move_keys)
+        query = Union(query, from_.select(move.from_location.as_('location'),
+                (-Sum(move.internal_quantity)).as_('quantity'),
+                *move_keys,
+                where=state_date_clause
+                & where
+                & move.from_location.in_(location_query)
+                & move.id.in_(move_rule_query)
+                & dest_clause_to,
+                group_by=[move.from_location] + move_keys),
+            all_=True)
+        if PeriodCache:
+            period_keys = [Column(period_cache, key).as_(key)
+                for key in grouping]
+            query = Union(query, from_period.select(
+                    period_cache.location.as_('location'),
+                    period_cache.internal_quantity.as_('quantity'),
+                    *period_keys,
+                    where=(period_cache.period
+                        == (period.id if period else None))
+                    & where_period
+                    & period_cache.location.in_(location_query)
+                    & dest_clause_period),
+                all_=True)
+        query_keys = [Column(query, key).as_(key) for key in grouping]
+        columns = ([query.location.as_('location')]
+            + query_keys
+            + [Sum(query.quantity).as_('quantity')])
+        query = query.select(*columns,
+            group_by=[query.location] + query_keys)
+        cursor.execute(*query)
         raw_lines = cursor.fetchall()
 
-        res_product_ids = set(product for _, product, _ in raw_lines)
-        res = dict(((location, product), quantity)
-                for location, product, quantity in raw_lines)
+        product_getter = itemgetter(grouping.index('product') + 1)
+        res_product_ids = set()
+        res = {}
+        keys = set()
+        for line in raw_lines:
+            location = line[0]
+            key = tuple(line[1:-1])
+            quantity = line[-1]
+            res[(location,) + key] = quantity
+            res_product_ids.add(product_getter(line))
+            keys.add(key)
 
         # Propagate quantities on from child locations to their parents
         if with_childs:
@@ -514,9 +426,10 @@ class Product:
                     locations.remove(l)
                     if l not in parent:
                         continue
-                    for product in res_product_ids:
-                        res.setdefault((parent[l], product), 0)
-                        res[(parent[l], product)] += res.get((l, product), 0)
+                    for key in keys:
+                        parent_key = (parent[l],) + key
+                        res.setdefault(parent_key, 0)
+                        res[parent_key] += res.get((l,) + key, 0)
                 next_leafs = set(locations)
                 for l in locations:
                     if l not in parent:
@@ -526,30 +439,20 @@ class Product:
                 leafs = next_leafs
 
             # clean result
-            for location, product in res.keys():
+            for key in res.keys():
+                location = key[0]
                 if location not in location_ids:
-                    del res[(location, product)]
+                    del res[key]
 
         # Round quantities
         default_uom = dict((p.id, p.default_uom) for p in
             cls.browse(list(res_product_ids)))
         for key, quantity in res.iteritems():
-            location, product = key
+            location = key[0]
+            product = product_getter(key)
             uom = default_uom[product]
             res[key] = Uom.round(quantity, uom.rounding)
 
-        # Complete result with missing products if asked
-        if not skip_zero:
-            # Search for all products, even if not linked with moves
-            if product_ids:
-                all_product_ids = product_ids
-            else:
-                all_product_ids = cls.search([])
-            keys = ((l, p) for l in location_ids for p in all_product_ids)
-            for location_id, product_id in keys:
-                if (location_id, product_id) not in res:
-                    res[(location_id, product_id)] = 0.0
-
         if wh_to_add:
             if product_ids is None:
                 product_ids = set((p for s, p in res))
@@ -634,24 +537,27 @@ class ProductQuantitiesByWarehouse(ModelSQL, ModelView):
         pool = Pool()
         Move = pool.get('stock.move')
         Location = pool.get('stock.location')
+        move = Move.__table__()
 
         product_id = Transaction().context.get('product')
         warehouse_id = Transaction().context.get('warehouse', -1)
-        warehouse_clause, warehouse_params = Location.search([
+        warehouse_query = Location.search([
                 ('parent', 'child_of', [warehouse_id]),
-                ], query_string=True, order=[])
-        return ('SELECT MAX(id) AS id, '
-                '0 AS create_uid, '
-                'NOW() AS create_date, '
-                'NULL AS write_uid, '
-                'NULL AS write_date, '
-                'COALESCE(effective_date, planned_date) AS date '
-            'FROM "' + Move._table + '" '
-            'WHERE product = %s '
-                'AND (from_location IN (' + warehouse_clause + ') '
-                    'OR to_location IN (' + warehouse_clause + ')) '
-                'AND COALESCE(effective_date, planned_date) IS NOT NULL '
-            'GROUP BY date, product', [product_id] + 2 * warehouse_params)
+                ], query=True, order=[])
+        date_column = Coalesce(move.effective_date, move.planned_date
+            ).as_('date')
+        return move.select(
+            Max(move.id).as_('id'),
+            Literal(0).as_('create_uid'),
+            Now().as_('create_date'),
+            Literal(None).as_('write_uid'),
+            Literal(None).as_('write_date'),
+            date_column,
+            where=(move.product == product_id)
+            & (move.from_location.in_(warehouse_query)
+                | move.to_location.in_(warehouse_query))
+            & (Coalesce(move.effective_date, move.planned_date) != None),
+            group_by=(date_column, move.product))
 
     @classmethod
     def get_quantity(cls, lines, name):
@@ -671,8 +577,8 @@ class ProductQuantitiesByWarehouse(ModelSQL, ModelView):
                 }
             with Transaction().set_context(**context):
                 quantities[date] = Product.products_by_location(
-                    [warehouse_id], [product_id], with_childs=True,
-                    skip_zero=False)[(warehouse_id, product_id)]
+                    [warehouse_id], [product_id],
+                    with_childs=True).get((warehouse_id, product_id), 0)
             try:
                 date_start = date + datetime.timedelta(1)
             except OverflowError:
